import os
import re
import sys
from pathlib import Path
import yaml
from typing import Any
from dataclasses import dataclass, field

from string import Template
from .template_classes.base import BaseTemplate
from .registry import get_template
from .templaters.base import ABCTemplater
from .templaters.registry import get_templater

template_lib_dir = Path(__file__).parent / 'template_lib'
TEMPLATE_PROPERTIES_FILENAME = 'template_properties.yaml'
DEFAULT_TEMPLATER = os.environ.get('skaf_TEMPLATER', 'jinja2')


@dataclass
class ScaffoldContext:
    project_name: str
    template_name: str
    output_dir: Path
    force: bool = False
    auto_use_defaults: bool | None = None
    project_path: Path = None
    template: BaseTemplate = None
    templater: ABCTemplater = None

    def __post_init__(self):
        self.project_name = sanitize_project_name(self.project_name)
        self.project_path = self.output_dir / self.project_name
        if not self.template and self.template_name:
            self.template = get_template(self.template_name)
        elif not self.template:
            raise ValueError("Either template or template_name must be provided.")
        self.templater = get_templater(self.template.properties.get('templater', DEFAULT_TEMPLATER))
        if self.auto_use_defaults is None:
            self.auto_use_defaults = self.template.properties.get('auto_use_defaults', False)


custom_var_type_mapper = {
    'str': str,
    'int': int,
    'float': float,
    'bool': bool,
    'list': lambda x: [x.strip() for x in x.split(',')],
    'dict': lambda x: dict(item.split('=') for item in x.split(',')),
}


def apply_templating(document: str,
                     variables: dict[str, Any],
                     templater: ABCTemplater,
                     document_filename: str = None
                     ) -> str:
    """
    Applies templating to the given document string using the provided variables
    using the specified templater.
    """
    try:
        return templater.render(document, variables, template_filename=document_filename)
    except KeyError as e:
        raise ValueError(f"Missing variable for templating: {e}")
    except Exception as e:
        raise RuntimeError(f"Error applying templating: {e}")


def get_template_dir(template_name: str) -> Path:
    """
    Returns the path to the specified template directory.
    If the template does not exist, raises a FileNotFoundError.
    """
    template_path = template_lib_dir / template_name
    if not template_path.exists():
        raise FileNotFoundError(f"Template '{template_name}' does not exist.")
    return template_path


def load_template_properties(template_name) -> dict:
    template_dir = get_template_dir(template_name)
    properties_filename = template_dir / TEMPLATE_PROPERTIES_FILENAME
    with open(properties_filename, 'r') as file:
        properties = yaml.safe_load(file)
    return properties


def add_project_name_variables(project_name: str, variables):
    """
    Adds project name and several other derivatives (project_name_snake, project_name_pascal, project_name_kebab)"""
    variables['project_name'] = project_name
    project_name_snake = re.sub(r'(?<!^)(?=[A-Z])', '_', project_name).lower()
    variables['project_name_snake'] = project_name_snake
    variables['project_name_pascal'] = ''.join(word.capitalize() for word in project_name_snake.split('_'))
    variables['project_name_kebab'] = project_name.replace('_', '-')
    variables['project_name_title'] = project_name.replace("_", " ").title()
    return variables


def get_template_variable_values(context: ScaffoldContext) -> dict[str, Any]:
    values = {}
    for custom_var in context.template.custom_variables:
        varname = custom_var['name']
        vartype = custom_var.get('type', 'str')
        caster = custom_var_type_mapper.get(vartype, str)
        default = custom_var.get('default')
        if context.auto_use_defaults and default is not None:
            try:
                values[varname] = caster(default)
            except ValueError as e:
                print(f"Default value for {varname}, {default} cannot be used with caster {caster}: {e}")
                sys.exit(1)
        else:
            defaultstr = f" [{default}]" if default else ""
            val = input(f"Enter value for {varname} ({vartype}){defaultstr}: ")
            if not val and default is not None:
                val = default
            try:
                values[varname] = caster(val)
            except ValueError as e:
                print(f"Invalid value for {varname} with type {vartype} and caster {caster}: {e}")
                sys.exit(1)
    values = add_project_name_variables(context.project_name, values)
    return values


def sanitize_project_name(name: str) -> str:
    """
    Converts the provided project name to a valid Python package name.
    Lowercases the name, replaces non-alphanumeric characters with underscores,
    and prepends an underscore if the name starts with a digit.
    """
    name = name.lower()
    sanitized = re.sub(r'\W+', '_', name)
    if sanitized and sanitized[0].isdigit():
        sanitized = "_" + sanitized
    return sanitized


def map_paths(context: ScaffoldContext,
              variables: dict[str, Any]
              ) -> dict[Path, str]:
    """
    Using the `template.documents` iterator method, get the `relpath, content` pairs
    and create a new `target_path, content` mapping. Perform templating on the template
    relpath using the provided variables.
    """
    targets = {}
    for relpath, content in context.template.documents():
        relpath = Path(apply_templating(relpath, variables, context.templater))
        target_path = context.project_path / relpath
        target_path.parent.mkdir(parents=True, exist_ok=True)
        targets[relpath] = content
    return targets


def scaffold_project(project_name: str,
                     template_name: str = None,
                     output_dir: str = None,
                     force: bool = False,
                     template: BaseTemplate = None,
                     auto_use_defaults: bool = True,
                     ) -> None:
    """
    Scaffold a new project based on the provided template and variables.
    Copies files from the template directory to the new project directory,
    replacing placeholders with the provided variable values.
    """
    if output_dir is None:
        output_dir = os.getcwd()
    output_dir = Path(output_dir)

    context = ScaffoldContext(
        project_name=project_name,
        template_name=template_name,
        output_dir=output_dir,
        force=force,
        auto_use_defaults=auto_use_defaults,
        template=template
    )

    variables = get_template_variable_values(context)

    path_mapping: dict[Path, str] = map_paths(
        context,
        variables
    )

    if not context.force:
        if all([
            context.project_path.exists(),
            context.project_path.is_dir(),
            os.listdir(context.project_path)
        ]):
            print(f"Project directory '{context.project_path}' already exists. Set --force to overwrite.")
            sys.exit(1)

    context.project_path.mkdir(parents=True, exist_ok=True)

    for target_path, content in path_mapping.items():
        target_filename = target_path.name
        content = apply_templating(
            content,
            variables,
            context.templater,
            target_filename
        )
        if target_path.suffix == context.templater.suffix:
            target_path = Path(target_path.stem)
        write_path = context.project_path / target_path
        write_path.parent.mkdir(parents=True, exist_ok=True)
        with open(write_path, 'w') as file:
            file.write(content)



import os
import pytest
from pathlib import Path

from skaf.template_classes.filesystem_template import FilesystemTemplate


class TestFilesystemTemplate:
    def test_template_init(self, sample_template_dir):
        template = FilesystemTemplate("test_template", str(sample_template_dir))
        assert template.template_name == "test_template"
        assert template.template_dir == str(sample_template_dir)
        assert template.properties is not None
        assert template.properties.get("templater") == "jinja2"
        assert len(template.properties.get("custom_variables", [])) == 2
    
    def test_load_properties(self, filesystem_template):
        properties = filesystem_template.properties
        assert properties is not None
        assert properties.get("templater") == "jinja2"
        assert len(properties.get("custom_variables", [])) == 2
        
        # Check custom variables
        vars = properties.get("custom_variables", [])
        assert vars[0].get("name") == "author"
        assert vars[0].get("default") == "Test Author"
        assert vars[1].get("name") == "version"
        assert vars[1].get("default") == "0.1.0"
    
    def test_missing_properties_file(self, temp_dir):
        # Create a template dir without a properties file
        bad_template_dir = temp_dir / "bad_template"
        bad_template_dir.mkdir()
        (bad_template_dir / "template").mkdir()
        
        with pytest.raises(FileNotFoundError):
            FilesystemTemplate("bad_template", str(bad_template_dir))
    
    def test_documents_iterator(self, filesystem_template):
        documents = list(filesystem_template.documents())
        assert len(documents) == 4  # We have 4 files in our fixture
        
        # Check that all expected files are included
        paths = [rel_path for rel_path, _ in documents]
        assert "pyproject.toml.jinja" in paths
        assert "README.md.jinja" in paths
        assert "src/{{ project_name }}/__init__.py" in paths
        assert "src/{{ project_name }}/main.py" in paths
        
        # Check content of a specific file
        for rel_path, content in documents:
            if rel_path == "README.md.jinja":
                assert "# {{ project_name }}" in content
                assert "A project by {{ author }}." in content
    
    def test_missing_template_directory(self, temp_dir):
        # Create a template dir without a template subdirectory
        bad_template_dir = temp_dir / "bad_template_dir"
        bad_template_dir.mkdir()
        
        # Add a properties file
        with open(bad_template_dir / "template_properties.yaml", "w") as f:
            f.write("templater: jinja2\n")
        
        template = FilesystemTemplate("bad_template", str(bad_template_dir))
        
        # Should raise when trying to iterate over documents
        with pytest.raises(FileNotFoundError):
            list(template.documents())
    
    def test_custom_variables_property(self, filesystem_template):
        custom_vars = filesystem_template.custom_variables
        assert len(custom_vars) == 2
        assert custom_vars[0]["name"] == "author"
        assert custom_vars[1]["name"] == "version"



import pytest
from unittest.mock import patch, MagicMock

from skaf.templaters.registry import get_templater, _templaters
from skaf.templaters.jinja import Jinja2Templater
from skaf.templaters.pystring import PystringTemplater


class TestTemplaterRegistryAdditional:
    
    def test_templater_registry_contents(self):
        # Verify templater registry contents
        assert 'jinja2' in _templaters
        assert 'pystring' in _templaters
        assert _templaters['jinja2'] == Jinja2Templater
        assert _templaters['pystring'] == PystringTemplater
    
    def test_get_templater_nonexistent_with_exception(self):
        # Call with nonexistent templater - should raise KeyError
        with pytest.raises(KeyError) as excinfo:
            get_templater('nonexistent_templater')
        
        assert "Templater 'nonexistent_templater' does not exist" in str(excinfo.value)
    
    @patch('skaf.templaters.registry._templaters')
    def test_get_templater_with_instantiation_error(self, mock_templaters):
        # Setup mock to raise exception during instantiation
        mock_templater_class = MagicMock(side_effect=ValueError('Templater initialization error'))
        mock_templaters.__getitem__.return_value = mock_templater_class
        
        # Call function and check for RuntimeError
        with pytest.raises(RuntimeError) as excinfo:
            get_templater('error_templater')
        
        assert "Error getting templater: ValueError: Templater initialization error" in str(excinfo.value)



from typing import Generator
import os
import yaml
from pathlib import Path

from .base import BaseTemplate, TemplateProperties


class FilesystemTemplate(BaseTemplate):

    template_properties_filename = 'template_properties.yaml'

    def __init__(self,
                 template_name: str,
                 template_dir: str,
                 ):
        self.template_dir = template_dir
        self.properties = self._load_properties()
        self.template_name = template_name

    def _load_properties(self) -> TemplateProperties:
        properties_filename = Path(self.template_dir) / Path(self.template_properties_filename)
        if not os.path.exists(properties_filename):
            raise FileNotFoundError(f"Template properties file '{properties_filename}' does not exist.")
        with open(properties_filename, 'r') as file:
            properties = yaml.safe_load(file)
        properties = properties or {}
        return properties

    def documents(self) -> Generator[tuple[str, str], None, None]:
        """
        Yields tuples of (relpath, content) for each document in the template.
        """
        template_root = Path(self.template_dir) / "template"
        if not os.path.exists(template_root):
            raise FileNotFoundError(f"Template root directory '{template_root}' does not exist.")
        for root, dirs, files in os.walk(template_root):
            rel_root = Path(root).relative_to(template_root)
            for name in files:
                rel_path_template = rel_root / name
                abs_path_template = template_root / rel_path_template
                with open(abs_path_template, 'r') as file:
                    content = file.read()
                yield str(rel_path_template), content
                



import pytest
import os
from unittest.mock import patch, MagicMock
import sys
from pathlib import Path
from skaf.cli import get_args, get_filesystem_template, main


class TestCliArgs:
    
    @patch('argparse.ArgumentParser.parse_args')
    def test_get_args(self, mock_parse_args):
        # Setup mock return value
        mock_args = MagicMock()
        mock_args.name = 'test_project'
        mock_args.template = 'test_template'
        mock_args.path = None
        mock_args.output = '/test/output'
        mock_args.force = False
        mock_args.confirm_defaults = True
        mock_args.debug = False
        mock_parse_args.return_value = mock_args
        mock_args.git = None
        
        # Call function
        args = get_args()
        
        # Verify results
        assert args.name == 'test_project'
        assert args.template == 'test_template'
        assert args.output == '/test/output'
        assert args.force is False
        assert args.confirm_defaults is True
        assert args.debug is False


class TestGetTemplate:
    
    @patch('os.path.isdir')
    @patch('skaf.cli.FilesystemTemplate')
    def test_get_template_valid_path(self, mock_template_class, mock_isdir):
        # Setup mocks
        mock_isdir.return_value = True
        mock_template = MagicMock()
        mock_template.template_name = 'test_template'
        mock_template_class.return_value = mock_template
        
        # Call function
        template = get_filesystem_template('/path/to/test_template')
        
        # Verify results
        mock_isdir.assert_called_once_with('/path/to/test_template')
        mock_template_class.assert_called_once_with('test_template', '/path/to/test_template')
        assert template == mock_template
    
    @patch('os.path.isdir')
    def test_get_template_invalid_path(self, mock_isdir):
        # Setup mock
        mock_isdir.return_value = False
        
        # Call function and verify exception
        with pytest.raises(ValueError) as excinfo:
            get_filesystem_template('/invalid/path')
        
        # Verify error message
        assert "Template path '/invalid/path' is not a valid directory" in str(excinfo.value)


class TestMain:
    
    @patch('skaf.cli.get_args')
    @patch('skaf.cli.get_filesystem_template')
    @patch('skaf.cli.scaffold_project')
    def test_main_success(self, mock_scaffold, mock_get_template, mock_get_args):
        # Setup mocks
        mock_args = MagicMock()
        mock_args.name = 'test_project'
        mock_args.template = 'test_template'
        mock_args.output = '/test/output'
        mock_args.force = False
        mock_args.path = None
        mock_args.auto_use_defaults = None
        mock_args.debug = False
        mock_get_args.return_value = mock_args
        mock_args.git = None
        
        # Call function
        with patch('builtins.print') as mock_print:
            main()
        
        # Verify results
        mock_scaffold.assert_called_once_with(
            project_name='test_project',
            template_name='test_template',
            output_dir='/test/output',
            force=False,
            template=None,
            auto_use_defaults=None
        )
        mock_print.assert_called_once_with(
            "Project 'test_project' initialized successfully using the 'test_template' template."
        )
    
    @patch('skaf.cli.get_args')
    @patch('skaf.cli.get_filesystem_template')
    @patch('skaf.cli.scaffold_project')
    def test_main_with_template_path(self, mock_scaffold, mock_get_template, mock_get_args):
        # Setup mocks
        mock_args = MagicMock()
        mock_args.name = 'test_project'
        mock_args.template = None
        mock_args.output = '/test/output'
        mock_args.force = False
        mock_args.path = '/path/to/template'
        mock_args.auto_use_defaults = True
        mock_args.debug = False
        mock_args.git = None
        mock_get_args.return_value = mock_args
        
        mock_template = MagicMock()
        mock_template.template_name = 'custom_template'
        mock_get_template.return_value = mock_template
        
        # Call function
        with patch('builtins.print') as mock_print:
            main()
        
        # Verify results
        mock_get_template.assert_called_once_with('/path/to/template')
        mock_scaffold.assert_called_once_with(
            project_name='test_project',
            template_name='custom_template',
            output_dir='/test/output',
            force=False,
            template=mock_template,
            auto_use_defaults=True
        )
        mock_print.assert_called_once_with(
            "Project 'test_project' initialized successfully using the 'custom_template' template."
        )
    
    @patch('skaf.cli.get_args')
    @patch('skaf.cli.scaffold_project')
    def test_main_with_error(self, mock_scaffold, mock_get_args):
        # Setup mocks
        mock_args = MagicMock()
        mock_args.name = 'test_project'
        mock_args.template = 'test_template'
        mock_args.output = '/test/output'
        mock_args.force = False
        mock_args.path = None
        mock_args.confirm_defaults = False
        mock_args.debug = False
        mock_get_args.return_value = mock_args
        mock_args.git = None
        
        mock_scaffold.side_effect = ValueError("Test error")
        
        # Call function
        with patch('builtins.print') as mock_print:
            with patch('sys.exit') as mock_exit:
                main()
        
        # Verify results
        mock_print.assert_called_once_with(
            "An error occurred while initializing the project: ValueError: Test error"
        )
        mock_exit.assert_called_once_with(1)
    
    @patch('skaf.cli.get_args')
    @patch('skaf.cli.scaffold_project')
    def test_main_debug_mode(self, mock_scaffold, mock_get_args):
        # Setup mocks
        mock_args = MagicMock()
        mock_args.name = 'test_project'
        mock_args.template = 'test_template'
        mock_args.output = '/test/output'
        mock_args.force = False
        mock_args.path = None
        mock_args.confirm_defaults = False
        mock_args.debug = True
        mock_get_args.return_value = mock_args
        mock_args.git = None
        
        mock_scaffold.side_effect = ValueError("Test error")
        
        # Call function and verify exception is re-raised
        with pytest.raises(ValueError) as excinfo:
            main()
        
        assert "Test error" in str(excinfo.value)



from .base import ABCTemplater
from .pystring import PystringTemplater
from .jinja import Jinja2Templater


_templaters: dict[str, ABCTemplater] = {
    "pystring": PystringTemplater,
    "jinja2": Jinja2Templater,
}


def get_templater(template_name: str, **kwargs) -> ABCTemplater:
    """
    Returns the templater instance for the given template name.
    If the template does not exist, raises a KeyError.
    """
    try:
        return _templaters[template_name](**kwargs)
    except KeyError:
        raise KeyError(f"Templater '{template_name}' does not exist.")
    except Exception as e:
        etype = type(e).__name__
        raise RuntimeError(f"Error getting templater: {etype}: {e}")



from importlib.metadata import entry_points
from pathlib import Path

from .template_classes.base import BaseTemplate
from .template_classes.filesystem_template import FilesystemTemplate


_entry_points_group = "skaf.template"
template_lib_dir = Path(__file__).parent / 'template_lib'


class LoadTemplateError(Exception):
    """
    Exception raised when a template cannot be loaded.
    """
    pass


class RegisterTemplateError(Exception):
    """
    Exception raised when a template cannot be registered.
    """
    pass


_templates = {}


def load_and_register_template_plugins():
    template_entry_points = entry_points(group=_entry_points_group)
    for tep in template_entry_points:
        _template = tep.load()
        register_template(_template)


def load_and_register_packaged_templates():
    """
    Loads and registers all packaged templates.
    """
    for template_dir in template_lib_dir.iterdir():
        if template_dir.is_dir():
            template_name = template_dir.name
            try:
                template = FilesystemTemplate(template_name, template_dir)
                register_template(template)
            except Exception as e:
                etype = type(e).__name__
                raise LoadTemplateError(f"Error loading '{template_name}': {etype}: {e}")


def register_template(template: BaseTemplate):
    """
    Registers a template class with the registry.
    The template class must inherit from BaseTemplate.
    """

    if not isinstance(template, BaseTemplate):
        raise RegisterTemplateError("Template must be a subclass of BaseTemplate.")
    
    template_name = template.template_name
    if template_name == "none":
        return

    if template_name in _templates:
        raise RegisterTemplateError(f"Template '{template_name}' is already registered.")
    
    _templates[template_name] = template


def get_template(template_name: str) -> BaseTemplate:
    """
    Returns the template class registered with the given name.
    If no template is found, raises a KeyError.
    """
    if template_name not in _templates:
        raise LoadTemplateError(f"Template '{template_name}' not found.")
    
    return _templates[template_name]
